(**********************************************************************)
(** - Copyright (C) 1991 Lehrstuhl III fuer Informatik, RWTH Aachen - *)
(** This library is free software; you can redistribute it and/or     *)
(** modify it under the terms of the GNU Library General Public       *)
(** License as published by the Free Software Foundation; either      *)
(** version 2 of the License, or (at your option) any later version.  *)
(** See the file COPYING.LIB for details.                             *)
(**                                                                   *)
(** Author: Peter Klein                                               *)
(**********************************************************************)
UNSAFE MODULE M2Dirs;

IMPORT Word, M3toC, Ustat, Unix, Ctypes;


PROCEDURE Access (name: TEXT; mode: AccessSet): INTEGER =
  VAR
    UnixFileName          := M3toC.CopyTtoS(name);
    umode                 := 0;
    Result      : INTEGER;

  BEGIN
    IF (Accessmode.accessX IN mode) THEN INC(umode, Unix.X_OK); END;
    IF (Accessmode.accessW IN mode) THEN INC(umode, Unix.W_OK); END;
    IF (Accessmode.accessR IN mode) THEN INC(umode, Unix.R_OK); END;

    Result := Unix.access(UnixFileName, umode);
    M3toC.FreeS(UnixFileName);
    RETURN Result;
  END Access;


PROCEDURE Chdir (name: TEXT): INTEGER =
  VAR
    UnixFileName          := M3toC.CopyTtoS(name);
    Result      : INTEGER;

  BEGIN
    Result := Unix.chdir(UnixFileName);
    M3toC.FreeS(UnixFileName);
    RETURN Result;
  END Chdir;


PROCEDURE Chmod (name: TEXT; mode: PermissionSet): INTEGER =
  VAR
    UnixFileName          := M3toC.CopyTtoS(name);
    Result      : INTEGER;

  BEGIN
    Result := Unix.chmod(UnixFileName, PermSetToMode(mode));
    M3toC.FreeS(UnixFileName);
    RETURN Result;
  END Chmod;


PROCEDURE Link (name1, name2: TEXT): INTEGER =
  VAR
    UnixFileName1          := M3toC.CopyTtoS(name1);
    UnixFileName2          := M3toC.CopyTtoS(name2);
    Result       : INTEGER;

  BEGIN
    Result := Unix.link(UnixFileName1, UnixFileName2);
    M3toC.FreeS(UnixFileName1);
    M3toC.FreeS(UnixFileName2);
    RETURN Result;
  END Link;


PROCEDURE Mkdir (name: TEXT; mode: PermissionSet): INTEGER =
  VAR
    UnixFileName          := M3toC.CopyTtoS(name);
    Result      : INTEGER;

  BEGIN
    Result := Unix.mkdir(UnixFileName, PermSetToMode(mode));
    M3toC.FreeS(UnixFileName);
    RETURN Result;
  END Mkdir;


PROCEDURE Rename (old, new: TEXT): INTEGER =
  VAR
    UnixFileName1          := M3toC.CopyTtoS(old);
    UnixFileName2          := M3toC.CopyTtoS(new);
    Result       : INTEGER;

  BEGIN
    Result := Unix.rename(UnixFileName1, UnixFileName2);
    M3toC.FreeS(UnixFileName1);
    M3toC.FreeS(UnixFileName2);
    RETURN Result;
  END Rename;


PROCEDURE Rmdir (name: TEXT): INTEGER =
  VAR
    UnixFileName          := M3toC.CopyTtoS(name);
    Result      : INTEGER;

  BEGIN
    Result := Unix.rmdir(UnixFileName);
    M3toC.FreeS(UnixFileName);
    RETURN Result;
  END Rmdir;


PROCEDURE Stat (name: TEXT; VAR buf: FStatType): INTEGER =
  VAR
    UnixFileName                         := M3toC.CopyTtoS(name);
    UnixStat    : Ustat.struct_stat_star;
    Result      : INTEGER;

  BEGIN
    UnixStat := NEW(Ustat.struct_stat_star);
    Result := Ustat.stat(UnixFileName, UnixStat);

    buf := FStatType{
             UnixStat.st_dev, UnixStat.st_ino,
             ModeToPermSet(UnixStat.st_mode), UnixStat.st_nlink,
             UnixStat.st_uid, UnixStat.st_gid, UnixStat.st_rdev,
             UnixStat.st_size, UnixStat.st_atime, UnixStat.st_spare1,
             UnixStat.st_mtime, UnixStat.st_spare2, UnixStat.st_ctime,
             UnixStat.st_spare3, UnixStat.st_blksize, UnixStat.st_blocks,
             ARRAY [0 .. 1] OF
               Ctypes.long{UnixStat.st_spare41, UnixStat.st_spare42},
             ModeToSFT(UnixStat.st_mode)};

    M3toC.FreeS(UnixFileName);
    DISPOSE(UnixStat);
    RETURN Result;
  END Stat;


PROCEDURE Unlink (name: TEXT): INTEGER =
  VAR
    UnixFileName          := M3toC.CopyTtoS(name);
    Result      : INTEGER;

  BEGIN
    Result := Unix.unlink(UnixFileName);
    M3toC.FreeS(UnixFileName);
    RETURN Result;
  END Unlink;


PROCEDURE PermSetToMode (S: PermissionSet): Ctypes.int =
  VAR Result := 0;

  BEGIN
    IF (Permission.permOX IN S) THEN INC(Result, Ustat.S_OEXEC); END;
    IF (Permission.permOW IN S) THEN INC(Result, Ustat.S_OWRITE); END;
    IF (Permission.permOR IN S) THEN INC(Result, Ustat.S_OREAD); END;

    IF (Permission.permGX IN S) THEN INC(Result, Ustat.S_GEXEC); END;
    IF (Permission.permGW IN S) THEN INC(Result, Ustat.S_GWRITE); END;
    IF (Permission.permGR IN S) THEN INC(Result, Ustat.S_GREAD); END;

    IF (Permission.permUX IN S) THEN INC(Result, Ustat.S_IEXEC); END;
    IF (Permission.permUW IN S) THEN INC(Result, Ustat.S_IWRITE); END;
    IF (Permission.permUR IN S) THEN INC(Result, Ustat.S_IREAD); END;

    IF (Permission.permSGID IN S) THEN INC(Result, Ustat.S_ISGID); END;
    IF (Permission.permSUID IN S) THEN INC(Result, Ustat.S_ISUID); END;
    RETURN Result;
  END PermSetToMode;


PROCEDURE ModeToPermSet (mode: Ctypes.int): PermissionSet =
  VAR Result := PermissionSet{};

  BEGIN
    IF (Word.And(mode, Ustat.S_OEXEC) # 0) THEN
      Result := Result + PermissionSet{Permission.permOX};
    END;
    IF (Word.And(mode, Ustat.S_OWRITE) # 0) THEN
      Result := Result + PermissionSet{Permission.permOW};
    END;
    IF (Word.And(mode, Ustat.S_OREAD) # 0) THEN
      Result := Result + PermissionSet{Permission.permOR};
    END;

    IF (Word.And(mode, Ustat.S_GEXEC) # 0) THEN
      Result := Result + PermissionSet{Permission.permGX};
    END;
    IF (Word.And(mode, Ustat.S_GWRITE) # 0) THEN
      Result := Result + PermissionSet{Permission.permGW};
    END;
    IF (Word.And(mode, Ustat.S_GREAD) # 0) THEN
      Result := Result + PermissionSet{Permission.permGR};
    END;

    IF (Word.And(mode, Ustat.S_IEXEC) # 0) THEN
      Result := Result + PermissionSet{Permission.permUX};
    END;
    IF (Word.And(mode, Ustat.S_IWRITE) # 0) THEN
      Result := Result + PermissionSet{Permission.permUW};
    END;
    IF (Word.And(mode, Ustat.S_IREAD) # 0) THEN
      Result := Result + PermissionSet{Permission.permUR};
    END;

    IF (Word.And(mode, Ustat.S_ISGID) # 0) THEN
      Result := Result + PermissionSet{Permission.permSGID};
    END;
    IF (Word.And(mode, Ustat.S_ISUID) # 0) THEN
      Result := Result + PermissionSet{Permission.permSUID};
    END;
    RETURN Result;
  END ModeToPermSet;


PROCEDURE ModeToSFT (mode: Ctypes.int): StatFileType =
  VAR Result: StatFileType;

  BEGIN
    IF (Word.And(mode, Ustat.S_IFREG) # 0) THEN
      RETURN StatFileType.statRegular;
    ELSIF (Word.And(mode, Ustat.S_IFDIR) # 0) THEN
      RETURN StatFileType.statDirectory;
    ELSIF (Word.And(mode, Ustat.S_IFCHR) # 0) THEN
      RETURN StatFileType.statCharacter;
    ELSIF (Word.And(mode, Ustat.S_IFBLK) # 0) THEN
      RETURN StatFileType.statBlock;
    ELSIF (Word.And(mode, Ustat.S_IFLNK) # 0) THEN
      RETURN StatFileType.statSymLink;
    ELSIF (Word.And(mode, Ustat.S_IFSOCK) # 0) THEN
      RETURN StatFileType.statSocket;
    END;
    RETURN Result;
  END ModeToSFT;

BEGIN
END M2Dirs.
