(**********************************************************************)
(** - Copyright (C) 1991 Lehrstuhl III fuer Informatik, RWTH Aachen - *)
(** This library is free software; you can redistribute it and/or     *)
(** modify it under the terms of the GNU Library General Public       *)
(** License as published by the Free Software Foundation; either      *)
(** version 2 of the License, or (at your option) any later version.  *)
(** See the file COPYING.LIB for details.                             *)
(**                                                                   *)
(** Author: Peter Klein                                               *)
(**********************************************************************)
UNSAFE MODULE M2Files;

IMPORT Word, Text, Ctypes, Cstdlib, M3toC, Unix, Uuio, Uerror;


REVEAL File = BRANDED REF FileInfo;


TYPE
  FileInfo = RECORD
               fdesc: Ctypes.int;
               fnam : TEXT;
               fstat: FileState;
             END;


PROCEDURE Open (             VAR file     : File;
                                 name     : TEXT;
                <* UNUSED *>     binText  : BinTextMode;
                                 writeMode: ReadWriteMode;
                             VAR state    : FileState      ) =
  VAR
    flags, result: Ctypes.int;
    UnixFileName              := M3toC.CopyTtoS(name);

  BEGIN
    CASE writeMode OF
      ReadWriteMode.readOnly => flags := Unix.O_RDONLY;
    | ReadWriteMode.readWrite => flags := Unix.O_RDWR;
    | ReadWriteMode.appendOnly => flags := Unix.O_APPEND + Unix.O_WRONLY;
    END;

    result := Unix.open(UnixFileName, flags, 0);
    M3toC.FreeS(UnixFileName);

    IF (result > 0) THEN
      file := NEW(File, fdesc := result, fnam := GetFullPathName(name),
                  fstat := FileState.ok);
      state := FileState.ok;
    ELSE
      state := ErrnoToFileState();
    END;
  END Open;


PROCEDURE Create (             VAR file    : File;
                                   name    : TEXT;
                  <* UNUSED *>     binText : BinTextMode;
                                   replMode: ReplaceMode;
                               VAR state   : FileState    ) =
  VAR
    flags, result: Ctypes.int;
    UnixFileName              := M3toC.CopyTtoS(name);

  BEGIN
    flags := Unix.O_RDWR + Unix.O_CREAT;
    IF (replMode = ReplaceMode.noReplace) THEN
      INC(flags, Unix.O_EXCL);
    END;

    result := Unix.open(UnixFileName, flags, Unix.MROWNER + Unix.MWOWNER);
    M3toC.FreeS(UnixFileName);

    IF (result > 0) THEN
      file := NEW(File, fdesc := result, fnam := GetFullPathName(name),
                  fstat := FileState.ok);
      state := FileState.ok;
    ELSE
      state := ErrnoToFileState();
    END;
  END Create;


PROCEDURE Close (VAR file: File; VAR state: FileState) =
  BEGIN
    IF (file = NIL) THEN
      state := FileState.notOpen;
    ELSE
      IF (Unix.close(file^.fdesc) = 0) THEN
        state := FileState.ok;
      ELSE
        state := ErrnoToFileState();
      END;
      file := NIL;
    END;
  END Close;


PROCEDURE Remove (VAR file: File; VAR state: FileState) =
  VAR UnixFileName: Ctypes.char_star;

  BEGIN
    IF (file = NIL) THEN
      state := FileState.notOpen;
    ELSE
      EVAL Unix.close(file^.fdesc);
      UnixFileName := M3toC.CopyTtoS(file^.fnam);
      IF (Unix.unlink(UnixFileName) = 0) THEN
        state := FileState.ok;
      ELSE
        state := ErrnoToFileState();
      END;
      M3toC.FreeS(UnixFileName);
      file := NIL;
    END;
  END Remove;


PROCEDURE Reset (file: File; VAR state: FileState) =
  BEGIN
    IF (file = NIL) THEN state := FileState.notOpen; RETURN; END;
    IF (file^.fstat = FileState.ok) THEN
      IF (Unix.lseek(file^.fdesc, 0, Unix.L_SET) # 0) THEN
        file^.fstat := ErrnoToFileState();
      ELSE
        file^.fstat := FileState.ok;
      END;
    END;
    state := file^.fstat;
  END Reset;


PROCEDURE Rewrite (file: File; VAR state: FileState) =
  BEGIN
    IF (file = NIL) THEN state := FileState.notOpen; RETURN; END;
    IF (file^.fstat = FileState.ok) THEN
      IF (Unix.lseek(file^.fdesc, 0, Unix.L_SET) # 0) THEN
        file^.fstat := ErrnoToFileState();
      ELSE
        IF (Unix.ftruncate(file^.fdesc, 0) # 0) THEN
          file^.fstat := ErrnoToFileState();
        ELSE
          file^.fstat := FileState.ok;
        END;
      END;
    END;
    state := file^.fstat;
  END Rewrite;


PROCEDURE Truncate (file: File; VAR state: FileState) =
  VAR Current: CARDINAL;

  BEGIN
    IF (file = NIL) THEN state := FileState.notOpen; RETURN; END;
    IF (file^.fstat = FileState.ok) THEN
      Current := Unix.lseek(file^.fdesc, 0, Unix.L_INCR);
      IF (Unix.ftruncate(file^.fdesc, Current) # 0) THEN
        file^.fstat := ErrnoToFileState();
      ELSE
        file^.fstat := FileState.ok;
      END;
    END;
    state := file^.fstat;
  END Truncate;


PROCEDURE Flush (file: File; VAR state: FileState) =
  BEGIN
    state := file^.fstat;
  END Flush;


PROCEDURE EOF (file: File): BOOLEAN =
  VAR Current, End: CARDINAL;

  BEGIN
    IF (file = NIL) THEN
      RETURN TRUE;
    ELSE
      Current := Unix.lseek(file^.fdesc, 0, Unix.L_INCR);
      End := Unix.lseek(file^.fdesc, 0, Unix.L_XTND);
      Current := Unix.lseek(file^.fdesc, Current, Unix.L_SET);
      RETURN (Current = End);
    END;
  END EOF;


PROCEDURE State (file: File): FileState =
  BEGIN
    IF (file = NIL) THEN
      RETURN FileState.notOpen;
    ELSE
      RETURN file^.fstat;
    END;
  END State;


PROCEDURE ResetState (file: File; VAR state: FileState) =
  BEGIN
    IF (file = NIL) THEN
      state := FileState.notOpen;
    ELSE
      file^.fstat := FileState.ok;
    END;
    state := file^.fstat;
  END ResetState;


PROCEDURE GetFileName (file: File; VAR name: TEXT; VAR state: FileState) =
  BEGIN
    IF (file = NIL) THEN
      state := FileState.notOpen;
    ELSE
      name := file^.fnam;
      state := file^.fstat;
    END;
  END GetFileName;


PROCEDURE ReadByte (file: File; VAR byte: CHAR; VAR state: FileState) =
  VAR BytesRead: CARDINAL;

  BEGIN
    ReadBytes(file, ADR(byte), BYTESIZE(CHAR), BytesRead, state);
  END ReadByte;


PROCEDURE ReadWord (file: File; VAR word: Word.T; VAR state: FileState) =
  VAR BytesRead: CARDINAL;

  BEGIN
    ReadBytes(file, ADR(word), BYTESIZE(Word.T), BytesRead, state);
  END ReadWord;


PROCEDURE ReadBlock (    file : File;
                     VAR block: ARRAY OF CHAR;
                     VAR state: FileState      ) =
  VAR BytesRead: CARDINAL;

  BEGIN
    ReadBytes(file, ADR(block[0]), BYTESIZE(block), BytesRead, state);
  END ReadBlock;


PROCEDURE ReadBytes (    file     : File;
                         addr     : ADDRESS;
                         bytes    : CARDINAL;
                     VAR bytesRead: CARDINAL;
                     VAR state    : FileState ) =
  VAR RetCode: Ctypes.int;

  BEGIN
    IF (file = NIL) THEN state := FileState.notOpen; RETURN; END;

    IF (file^.fstat = FileState.ok) THEN
      RetCode := Uuio.read(file^.fdesc, addr, bytes);
      IF (RetCode <= 0) THEN
        IF (RetCode < 0) THEN
          file^.fstat := ErrnoToFileState();
        ELSE
          file^.fstat := FileState.endError;
        END;
      ELSE
        file^.fstat := FileState.ok;
        bytesRead := RetCode;
      END;
    END;
    state := file^.fstat;
  END ReadBytes;


PROCEDURE WriteByte (file: File; byte: CHAR; VAR state: FileState) =
  BEGIN
    WriteBytes(file, ADR(byte), BYTESIZE(CHAR), state);
  END WriteByte;


PROCEDURE WriteWord (file: File; word: Word.T; VAR state: FileState) =
  BEGIN
    WriteBytes(file, ADR(word), BYTESIZE(Word.T), state);
  END WriteWord;


PROCEDURE WriteBlock (         file : File;
                      READONLY block: ARRAY OF CHAR;
                      VAR      state: FileState      ) =
  BEGIN
    WriteBytes(file, ADR(block[0]), BYTESIZE(block), state);
  END WriteBlock;


PROCEDURE WriteBytes (    file     : File;
                          addr     : ADDRESS;
                          bytecount: CARDINAL;
                      VAR state    : FileState ) =
  VAR RetCode: Ctypes.int;

  BEGIN
    IF (file = NIL) THEN state := FileState.notOpen; RETURN; END;

    IF (file^.fstat = FileState.ok) THEN
      RetCode := Uuio.write(file^.fdesc, addr, bytecount);
      IF (RetCode < 0) THEN
        file^.fstat := ErrnoToFileState();
      ELSE
        file^.fstat := FileState.ok;
      END;
    END;
    state := file^.fstat;
  END WriteBytes;


PROCEDURE GetFilePos (file: File; VAR pos: FilePosition) =
  BEGIN
    pos := Unix.lseek(file^.fdesc, 0, Unix.L_INCR);
  END GetFilePos;


PROCEDURE GetEOF (file: File; VAR pos: FilePosition) =
  VAR Current: CARDINAL;

  BEGIN
    Current := Unix.lseek(file^.fdesc, 0, Unix.L_INCR);
    pos := Unix.lseek(file^.fdesc, 0, Unix.L_XTND);
    Current := Unix.lseek(file^.fdesc, Current, Unix.L_SET);
  END GetEOF;


PROCEDURE GetBOF (<* UNUSED *> file: File; VAR pos: FilePosition) =
  BEGIN
    pos := 0;
  END GetBOF;


PROCEDURE CalcFilePos (<* UNUSED *>     file         : File;
                                    VAR pos          : FilePosition;
                                        numOfElements: INTEGER;
                                        elementLength: CARDINAL      ) =
  BEGIN
    INC(pos, numOfElements * elementLength);
  END CalcFilePos;


PROCEDURE SetFilePos (file: File; pos: FilePosition; VAR state: FileState) =
  BEGIN
    IF (file = NIL) THEN state := FileState.notOpen; RETURN; END;

    IF (file^.fstat = FileState.ok) THEN
      pos := Unix.lseek(file^.fdesc, pos, Unix.L_SET);
      IF (pos < 0) THEN
        file^.fstat := ErrnoToFileState();
      ELSE
        file^.fstat := FileState.ok;
      END;
    END;
    state := file^.fstat;
  END SetFilePos;


PROCEDURE ErrnoToFileState (): FileState =
  VAR state: FileState;

  BEGIN
    CASE Uerror.errno OF
    | Uerror.ENAMETOOLONG, Uerror.ENOTDIR => state := FileState.nameError;
    | Uerror.ENOENT => state := FileState.noFile;
    | Uerror.EEXIST => state := FileState.existingFile;
    | Uerror.EIO, Uerror.EFAULT => state := FileState.deviceError;
    | Uerror.ENFILE, Uerror.EDQUOT, Uerror.EFBIG, Uerror.EMFILE,
        Uerror.EMLINK, Uerror.ENOMEM, Uerror.ENOSPC, Uerror.EUSERS =>
        state := FileState.noMoreRoom;
    | Uerror.EACCES, Uerror.EISDIR, Uerror.EBADF, Uerror.ENODEV,
        Uerror.ENOTEMPTY, Uerror.EPERM, Uerror.EROFS =>
        state := FileState.accessError;
    ELSE
      state := FileState.otherError;
    END;
    RETURN state;
  END ErrnoToFileState;


PROCEDURE GetFullPathName (rp: TEXT): TEXT =
  VAR
    UnixPath: Ctypes.char_star;
    Base    : TEXT;

  BEGIN
    IF (Text.GetChar(rp, 0) = '/') THEN
      RETURN rp;
    ELSE
      UnixPath := Cstdlib.malloc(Unix.MaxPathLen);
      UnixPath := Unix.getwd(UnixPath);
      Base := M3toC.CopyStoT(UnixPath);
      Cstdlib.free(UnixPath);
      RETURN Base & "/" & rp;
    END;
  END GetFullPathName;

BEGIN
END M2Files.
