(**********************************************************************)
(** - Copyright (C) 1991 Lehrstuhl III fuer Informatik, RWTH Aachen - *)
(** This library is free software; you can redistribute it and/or     *)
(** modify it under the terms of the GNU Library General Public       *)
(** License as published by the Free Software Foundation; either      *)
(** version 2 of the License, or (at your option) any later version.  *)
(** See the file COPYING.LIB for details.                             *)
(**                                                                   *)
(** Author of the generic parser: Frank Hoefer                        *)
(** For the Modula-2 to -3 converter: Peter Klein                     *)
(**********************************************************************)
MODULE Parser;

IMPORT Scanner;

FROM Scanner IMPORT TokenType;

IMPORT Wr, Stdio, Thread, Fmt;


EXCEPTION ParserError;


TYPE TokenSet = SET OF TokenType;


CONST
  Expression = TokenSet{
                 TokenType.Plus, TokenType.Minus, TokenType.Identifier,
                 TokenType.IntegerConst, TokenType.LongConst,
                 TokenType.StringConst, TokenType.CharConst,
                 TokenType.RealConst, TokenType.CLBracket,
                 TokenType.LBracket, TokenType.Not};

  AddOperator = TokenSet{TokenType.Plus, TokenType.Minus, TokenType.Or};

  MulOperator = TokenSet{TokenType.Mult, TokenType.Division, TokenType.Div,
                         TokenType.Mod, TokenType.Rem, TokenType.And,
                         TokenType.Ampersand};

  Relation = TokenSet{TokenType.Equal, TokenType.NEqual, TokenType.Less,
                      TokenType.LEqual, TokenType.Greater,
                      TokenType.GEqual, TokenType.In};


VAR fileName: TEXT;


PROCEDURE ParseElement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseExpression();
    IF (Scanner.lookahead = TokenType.Range) THEN
      Scanner.Match(TokenType.Range, TokenType.Range);
      ParseExpression();
    END;
  END ParseElement;


PROCEDURE ParseSet ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    IF (Scanner.lookahead = TokenType.Identifier) THEN
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      WHILE (Scanner.lookahead = TokenType.Period) DO
        Scanner.Match(TokenType.Period, TokenType.Period);
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      END;
    END;
    Scanner.Match(TokenType.CLBracket, TokenType.CLBracket);
    IF (Scanner.lookahead IN Expression) THEN
      ParseElement();
      WHILE (Scanner.lookahead = TokenType.Comma) DO
        Scanner.Match(TokenType.Comma, TokenType.Comma);
        ParseElement();
      END;
    END;
    Scanner.Match(TokenType.CRBracket, TokenType.CRBracket);
  END ParseSet;


PROCEDURE ParseConstElement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseConstExpression();
    IF (Scanner.lookahead = TokenType.Range) THEN
      Scanner.Match(TokenType.Range, TokenType.Range);
      ParseConstExpression();
    END;
  END ParseConstElement;


PROCEDURE ParseConstSet ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.CLBracket, TokenType.CLBracket);
    IF (Scanner.lookahead IN Expression) THEN
      ParseConstElement();
      WHILE (Scanner.lookahead = TokenType.Comma) DO
        Scanner.Match(TokenType.Comma, TokenType.Comma);
        ParseConstElement();
      END;
    END;
    Scanner.Match(TokenType.CRBracket, TokenType.CRBracket);
  END ParseConstSet;


PROCEDURE ParseConstFactor ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    CASE Scanner.lookahead OF
      TokenType.Not =>
        Scanner.Match(TokenType.Not, TokenType.Not);
        ParseConstFactor();
    | TokenType.LBracket =>
        Scanner.Match(TokenType.LBracket, TokenType.LBracket);
        ParseConstExpression();
        Scanner.Match(TokenType.RBracket, TokenType.RBracket);
    | TokenType.StringConst =>
        Scanner.Match(TokenType.StringConst, TokenType.StringConst);
    | TokenType.CharConst =>
        Scanner.Match(TokenType.CharConst, TokenType.CharConst);
    | TokenType.IntegerConst =>
        Scanner.Match(TokenType.IntegerConst, TokenType.IntegerConst);
    | TokenType.LongConst =>
        Scanner.Match(TokenType.LongConst, TokenType.LongConst);
    | TokenType.RealConst =>
        Scanner.Match(TokenType.RealConst, TokenType.RealConst);
    | TokenType.Identifier =>
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        WHILE (Scanner.lookahead = TokenType.Period) DO
          Scanner.Match(TokenType.Period, TokenType.Period);
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        END;
        IF (Scanner.lookahead = TokenType.CLBracket) THEN
          ParseConstSet();
        ELSIF (Scanner.lookahead = TokenType.LBracket) THEN
          ParseActualParameters();
        END;
    ELSE
      RAISE ParserError;
    END;
  END ParseConstFactor;


PROCEDURE ParseConstTerm ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseConstFactor();
    WHILE (Scanner.lookahead IN MulOperator) DO
      Scanner.Match(Scanner.lookahead, Scanner.lookahead);
      ParseConstFactor();
    END;
  END ParseConstTerm;


PROCEDURE ParseSimpleConstExpression ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    IF (Scanner.lookahead = TokenType.Plus) THEN
      Scanner.Match(TokenType.Plus, TokenType.Plus);
    ELSIF (Scanner.lookahead = TokenType.Minus) THEN
      Scanner.Match(TokenType.Minus, TokenType.Minus);
    END;
    ParseConstTerm();
    WHILE (Scanner.lookahead IN AddOperator) DO
      Scanner.Match(Scanner.lookahead, Scanner.lookahead);
      ParseConstTerm();
    END;
  END ParseSimpleConstExpression;


PROCEDURE ParseConstExpression ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseSimpleConstExpression();
    IF (Scanner.lookahead IN Relation) THEN
      Scanner.Match(Scanner.lookahead, Scanner.lookahead);
      ParseSimpleConstExpression();
    END;
  END ParseConstExpression;


PROCEDURE ParseExpression ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseSimpleExpression();
    IF (Scanner.lookahead IN Relation) THEN
      Scanner.Match(Scanner.lookahead, Scanner.lookahead);
      ParseSimpleExpression();
    END;
  END ParseExpression;


PROCEDURE ParseActualParameters ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.LBracket, TokenType.LBracket);
    IF (Scanner.lookahead IN Expression) THEN
      ParseExpression();
      WHILE (Scanner.lookahead = TokenType.Comma) DO
        Scanner.Match(TokenType.Comma, TokenType.Comma);
        ParseExpression();
      END;
    END;
    Scanner.Match(TokenType.RBracket, TokenType.RBracket);
  END ParseActualParameters;


PROCEDURE ParseCaseLabelList ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseCaseLabels();
    WHILE (Scanner.lookahead = TokenType.Comma) DO
      Scanner.Match(TokenType.Comma, TokenType.Comma);
      ParseCaseLabels();
    END;
  END ParseCaseLabelList;


PROCEDURE ParseCaseLabels ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseConstExpression();
    IF (Scanner.lookahead = TokenType.Range) THEN
      Scanner.Match(TokenType.Range, TokenType.Range);
      ParseConstExpression();
    END;
  END ParseCaseLabels;


PROCEDURE ParseDesignator ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    LOOP
      CASE Scanner.lookahead OF
        TokenType.Period =>
          Scanner.Match(TokenType.Period, TokenType.Period);
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      | TokenType.Arrow => Scanner.Match(TokenType.Arrow, TokenType.Arrow);
      | TokenType.ALBracket =>
          Scanner.Match(TokenType.ALBracket, TokenType.ALBracket);
          ParseExpression();
          WHILE (Scanner.lookahead = TokenType.Comma) DO
            Scanner.Match(TokenType.Comma, TokenType.Comma);
            ParseExpression();
          END;
          Scanner.Match(TokenType.ARBracket, TokenType.ARBracket);
      | TokenType.CLBracket =>
          Scanner.Match(TokenType.CLBracket, TokenType.CLBracket);
          IF (Scanner.lookahead IN Expression) THEN
            ParseElement();
            WHILE (Scanner.lookahead = TokenType.Comma) DO
              Scanner.Match(TokenType.Comma, TokenType.Comma);
              ParseElement();
            END;
          END;
          Scanner.Match(TokenType.CRBracket, TokenType.CRBracket);
      ELSE
        EXIT;
      END;
    END;
  END ParseDesignator;


PROCEDURE ParseFactor ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    CASE Scanner.lookahead OF
      TokenType.IntegerConst, TokenType.LongConst, TokenType.StringConst,
        TokenType.CharConst, TokenType.RealConst =>
        Scanner.Match(Scanner.lookahead, Scanner.lookahead);
    | TokenType.CLBracket => ParseSet();
    | TokenType.LBracket =>
        Scanner.Match(TokenType.LBracket, TokenType.LBracket);
        ParseExpression();
        Scanner.Match(TokenType.RBracket, TokenType.RBracket);
    | TokenType.Not =>
        Scanner.Match(TokenType.Not, TokenType.Not);
        ParseFactor();
    | TokenType.Identifier =>
        ParseDesignator();
        IF (Scanner.lookahead = TokenType.LBracket) THEN
          ParseActualParameters();
        END;
    ELSE
      RAISE ParserError;
    END;
  END ParseFactor;


PROCEDURE ParseTerm ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseFactor();
    WHILE (Scanner.lookahead IN MulOperator) DO
      Scanner.Match(Scanner.lookahead, Scanner.lookahead);
      ParseFactor();
    END;
  END ParseTerm;


PROCEDURE ParseSimpleExpression ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    IF (Scanner.lookahead = TokenType.Plus) THEN
      Scanner.Match(TokenType.Plus, TokenType.Plus);
      ParseTerm();
    ELSIF (Scanner.lookahead = TokenType.Minus) THEN
      Scanner.Match(TokenType.Minus, TokenType.Minus);
      ParseTerm();
    END;
    ParseTerm();
    WHILE (Scanner.lookahead IN AddOperator) DO
      Scanner.Match(Scanner.lookahead, Scanner.lookahead);
      ParseTerm();
    END;
  END ParseSimpleExpression;


PROCEDURE ParseStatementSequence ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    LOOP
      CASE Scanner.lookahead OF
        TokenType.Identifier =>
          ParseDesignator();
          CASE Scanner.lookahead OF
            TokenType.Assign =>
              Scanner.Match(TokenType.Assign, TokenType.Assign);
              ParseExpression();
          | TokenType.LBracket => ParseActualParameters();
          | TokenType.Semicolon, TokenType.End, TokenType.Until,
              TokenType.Bar, TokenType.Else, TokenType.Elsif =>
              TRY
                Wr.PutText(Stdio.stdout, "() ");
              EXCEPT
                Wr.Failure, Thread.Alerted =>
                  RAISE Scanner.IOError("error in writing output");
              END;
          ELSE
            RAISE ParserError;
          END;
      | TokenType.If => ParseIfStatement();
      | TokenType.Case => ParseCaseStatement();
      | TokenType.While => ParseWhileStatement();
      | TokenType.Repeat => ParseRepeatStatement();
      | TokenType.Loop => ParseLoopStatement();
      | TokenType.For => ParseForStatement();
      | TokenType.With => ParseWithStatement();
      | TokenType.Exit => Scanner.Match(TokenType.Exit, TokenType.Exit);
      | TokenType.Return =>
          Scanner.Match(TokenType.Return, TokenType.Return);
          IF (Scanner.lookahead IN Expression) THEN ParseExpression(); END;
      | TokenType.End, TokenType.Until, TokenType.Bar, TokenType.Else,
          TokenType.Elsif =>
      | TokenType.Semicolon =>
      ELSE
        RAISE ParserError;
      END;

      CASE Scanner.lookahead OF
        TokenType.End, TokenType.Until, TokenType.Bar, TokenType.Else,
          TokenType.Elsif =>
          EXIT;
      ELSE
        Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
      END;
    END;
  END ParseStatementSequence;


PROCEDURE ParseIfStatement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.If, TokenType.If);
    ParseExpression();
    Scanner.Match(TokenType.Then, TokenType.Then);
    ParseStatementSequence();
    WHILE (Scanner.lookahead = TokenType.Elsif) DO
      Scanner.Match(TokenType.Elsif, TokenType.Elsif);
      ParseExpression();
      Scanner.Match(TokenType.Then, TokenType.Then);
      ParseStatementSequence();
    END;
    IF (Scanner.lookahead = TokenType.Else) THEN
      Scanner.Match(TokenType.Else, TokenType.Else);
      ParseStatementSequence();
    END;
    Scanner.Match(TokenType.End, TokenType.End);
  END ParseIfStatement;


PROCEDURE ParseCase ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ParseCaseLabelList();
    Scanner.Match(TokenType.Colon, TokenType.ImpArrow);
    ParseStatementSequence();
  END ParseCase;


PROCEDURE ParseCaseStatement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Case, TokenType.Case);
    ParseExpression();
    Scanner.Match(TokenType.Of, TokenType.Of);

    IF (Scanner.lookahead IN Expression) THEN ParseCase(); END;

    WHILE (Scanner.lookahead = TokenType.Bar) DO
      Scanner.Match(TokenType.Bar, TokenType.Bar);
      IF (Scanner.lookahead IN Expression) THEN ParseCase(); END;
    END;

    IF (Scanner.lookahead = TokenType.Else) THEN
      Scanner.Match(TokenType.Else, TokenType.Else);
      ParseStatementSequence();
    END;

    Scanner.Match(TokenType.End, TokenType.End);
  END ParseCaseStatement;


PROCEDURE ParseWhileStatement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.While, TokenType.While);
    ParseExpression();
    Scanner.Match(TokenType.Do, TokenType.Do);
    ParseStatementSequence();
    Scanner.Match(TokenType.End, TokenType.End);
  END ParseWhileStatement;


PROCEDURE ParseRepeatStatement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Repeat, TokenType.Repeat);
    ParseStatementSequence();
    Scanner.Match(TokenType.Until, TokenType.Until);
    ParseExpression();
  END ParseRepeatStatement;


PROCEDURE ParseForStatement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.For, TokenType.For);
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    Scanner.Match(TokenType.Assign, TokenType.Assign);
    ParseExpression();
    Scanner.Match(TokenType.To, TokenType.To);
    ParseExpression();
    IF (Scanner.lookahead = TokenType.By) THEN
      Scanner.Match(TokenType.By, TokenType.By);
      ParseExpression();
    END;
    Scanner.Match(TokenType.Do, TokenType.Do);
    ParseStatementSequence();
    Scanner.Match(TokenType.End, TokenType.End);
  END ParseForStatement;


PROCEDURE ParseLoopStatement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Loop, TokenType.Loop);
    ParseStatementSequence();
    Scanner.Match(TokenType.End, TokenType.End);
  END ParseLoopStatement;


PROCEDURE ParseWithStatement ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    ErrorOut("warning: WITH statement");
    Scanner.Match(TokenType.With, TokenType.With);
    ParseDesignator();
    Scanner.Match(TokenType.Do, TokenType.Do);

    ParseStatementSequence();

    Scanner.Match(TokenType.End, TokenType.End);
  END ParseWithStatement;


PROCEDURE ParseSimpleType ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    CASE Scanner.lookahead OF
      TokenType.LBracket =>
        Scanner.Match(TokenType.LBracket, TokenType.CLBracket);
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        WHILE (Scanner.lookahead = TokenType.Comma) DO
          Scanner.Match(TokenType.Comma, TokenType.Comma);
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        END;
        Scanner.Match(TokenType.RBracket, TokenType.CRBracket);
    | TokenType.ALBracket =>
        Scanner.Match(TokenType.ALBracket, TokenType.ALBracket);
        ParseConstExpression();
        Scanner.Match(TokenType.Range, TokenType.Range);
        ParseConstExpression();
        Scanner.Match(TokenType.ARBracket, TokenType.ARBracket);
    | TokenType.Identifier =>
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        CASE Scanner.lookahead OF
          TokenType.Period =>
            WHILE (Scanner.lookahead = TokenType.Period) DO
              Scanner.Match(TokenType.Period, TokenType.Period);
              Scanner.Match(TokenType.Identifier, TokenType.Identifier);
            END;
        | TokenType.ALBracket =>
            Scanner.Match(TokenType.ALBracket, TokenType.ALBracket);
            ParseConstExpression();
            Scanner.Match(TokenType.Range, TokenType.Range);
            ParseConstExpression();
            Scanner.Match(TokenType.ARBracket, TokenType.ARBracket);
        ELSE
        END;
    ELSE
      RAISE ParserError;
    END;
  END ParseSimpleType;


PROCEDURE ParseVariant ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    IF (Scanner.lookahead IN Expression) THEN
      ParseCaseLabelList();
      Scanner.Match(TokenType.Colon, TokenType.Colon);
      ParseFieldListSequence();
    END;
  END ParseVariant;


PROCEDURE ParseFieldList ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    CASE Scanner.lookahead OF
      TokenType.Identifier =>
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);

        WHILE (Scanner.lookahead = TokenType.Comma) DO
          Scanner.Match(TokenType.Comma, TokenType.Comma);
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        END;
        Scanner.Match(TokenType.Colon, TokenType.Colon);
        ParseType();
    | TokenType.Case =>
        ErrorOut("warning: variant RECORD");
        Scanner.Match(TokenType.Case, TokenType.Case);
        IF (Scanner.lookahead = TokenType.Identifier) THEN
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        END;
        Scanner.Match(TokenType.Colon, TokenType.Colon);

        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        WHILE (Scanner.lookahead = TokenType.Period) DO
          Scanner.Match(TokenType.Period, TokenType.Period);
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
        END;

        Scanner.Match(TokenType.Of, TokenType.Of);
        ParseVariant();
        WHILE (Scanner.lookahead = TokenType.Bar) DO
          Scanner.Match(TokenType.Bar, TokenType.Bar);
          ParseVariant();
        END;
        IF (Scanner.lookahead = TokenType.Else) THEN
          Scanner.Match(TokenType.Else, TokenType.Else);
          ParseFieldListSequence();
        END;
        Scanner.Match(TokenType.End, TokenType.End);
    ELSE
      RAISE ParserError;
    END;
  END ParseFieldList;


PROCEDURE ParseFieldListSequence ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    IF ((Scanner.lookahead = TokenType.Identifier)
          OR (Scanner.lookahead = TokenType.Case)) THEN
      ParseFieldList();
      WHILE (Scanner.lookahead = TokenType.Semicolon) DO
        Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
        IF ((Scanner.lookahead = TokenType.Identifier)
              OR (Scanner.lookahead = TokenType.Case)) THEN
          ParseFieldList();
        END;
      END;
    END;
  END ParseFieldListSequence;


PROCEDURE ParseRecordType ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Record, TokenType.Record);
    ParseFieldListSequence();
    Scanner.Match(TokenType.End, TokenType.End);
  END ParseRecordType;


PROCEDURE ParseArrayType ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Array, TokenType.Array);
    ParseSimpleType();
    WHILE (Scanner.lookahead = TokenType.Comma) DO
      Scanner.Match(TokenType.Comma, TokenType.Comma);
      ParseSimpleType();
    END;
    Scanner.Match(TokenType.Of, TokenType.Of);
    ParseType();
  END ParseArrayType;


PROCEDURE ParseSetType ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Set, TokenType.Set);
    Scanner.Match(TokenType.Of, TokenType.Of);
    ParseType();
  END ParseSetType;


PROCEDURE ParsePointerType ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Pointer, TokenType.Branded);
    Scanner.Match(TokenType.To, TokenType.Ref);
    ParseType();
  END ParsePointerType;


PROCEDURE ParseProcedureType ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Procedure, TokenType.Procedure);
    IF (Scanner.lookahead = TokenType.LBracket) THEN
      ParseFormTypeList();
    ELSE
      TRY
        Wr.PutText(Stdio.stdout, "() ");
      EXCEPT
        Wr.Failure, Thread.Alerted =>
          RAISE Scanner.IOError("error in writing output");
      END;
    END;
    IF (Scanner.lookahead = TokenType.Colon) THEN
      Scanner.Match(TokenType.Colon, TokenType.Colon);
      ParseSimpleType();
    END;
  END ParseProcedureType;


PROCEDURE ParseType ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    CASE Scanner.lookahead OF
      TokenType.Array => ParseArrayType();
    | TokenType.Pointer => ParsePointerType();
    | TokenType.Set => ParseSetType();
    | TokenType.Record => ParseRecordType();
    | TokenType.Procedure => ParseProcedureType();
    ELSE
      ParseSimpleType();
    END;
  END ParseType;


PROCEDURE ParseFormTypeList ()
  RAISES {Scanner.IOError, Scanner.ScanError} =
  BEGIN
    Scanner.Match(TokenType.LBracket, TokenType.LBracket);
    IF ((Scanner.lookahead = TokenType.Array)
          OR (Scanner.lookahead = TokenType.Var)
          OR (Scanner.lookahead = TokenType.Identifier)) THEN
      IF (Scanner.lookahead = TokenType.Var) THEN
        Scanner.Match(TokenType.Var, TokenType.Var);
        ParseFormType();
      ELSE
        ParseFormType();
      END;
      WHILE (Scanner.lookahead = TokenType.Comma) DO
        Scanner.Match(TokenType.Comma, TokenType.Comma);
        IF (Scanner.lookahead = TokenType.Var) THEN
          Scanner.Match(TokenType.Var, TokenType.Var);
          ParseFormType();
        ELSE
          ParseFormType();
        END;
      END;
    END;
    Scanner.Match(TokenType.RBracket, TokenType.RBracket);
  END ParseFormTypeList;


PROCEDURE ParseBlock ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    WHILE ((Scanner.lookahead = TokenType.Const)
             OR (Scanner.lookahead = TokenType.Type)
             OR (Scanner.lookahead = TokenType.Var)
             OR (Scanner.lookahead = TokenType.Procedure)
             OR (Scanner.lookahead = TokenType.Module)) DO
      ParseDeclaration();
    END;
    IF (Scanner.lookahead = TokenType.Begin) THEN
      Scanner.Match(TokenType.Begin, TokenType.Begin);
    ELSE
      TRY
        Wr.PutText(Stdio.stdout, "\nBEGIN\n");
      EXCEPT
        Wr.Failure, Thread.Alerted =>
          RAISE Scanner.IOError("error in writing output");
      END;
    END;
    IF (Scanner.lookahead # TokenType.End) THEN
      ParseStatementSequence();
    END;
    Scanner.Match(TokenType.End, TokenType.End);
  END ParseBlock;


PROCEDURE ParseProcedureDeclaration ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Procedure, TokenType.Procedure);
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);

    IF (Scanner.lookahead = TokenType.LBracket) THEN
      ParseFormalParameters();
    ELSE
      TRY
        Wr.PutText(Stdio.stdout, "() ");
      EXCEPT
        Wr.Failure, Thread.Alerted =>
          RAISE Scanner.IOError("error in writing output");
      END;
    END;
    Scanner.Match(TokenType.Semicolon, TokenType.Equal);

    IF (Scanner.lookahead = TokenType.Forward) THEN
      Scanner.Match(TokenType.Forward, TokenType.Forward);
    ELSE
      ParseBlock();
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    END;
  END ParseProcedureDeclaration;


PROCEDURE ParseProcedureHeading ()
  RAISES {Scanner.IOError, Scanner.ScanError} =
  BEGIN
    Scanner.Match(TokenType.Procedure, TokenType.Procedure);
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    IF (Scanner.lookahead = TokenType.LBracket) THEN
      ParseFormalParameters();
    ELSE
      TRY
        Wr.PutText(Stdio.stdout, "() ");
      EXCEPT
        Wr.Failure, Thread.Alerted =>
          RAISE Scanner.IOError("error in writing output");
      END;
    END;
  END ParseProcedureHeading;


PROCEDURE ParseFormalParameters ()
  RAISES {Scanner.IOError, Scanner.ScanError} =
  BEGIN
    Scanner.Match(TokenType.LBracket, TokenType.LBracket);
    IF ((Scanner.lookahead = TokenType.Var)
          OR (Scanner.lookahead = TokenType.Identifier)) THEN
      ParseFPSection();
    END;
    WHILE (Scanner.lookahead = TokenType.Semicolon) DO
      Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
      ParseFPSection();
    END;
    Scanner.Match(TokenType.RBracket, TokenType.RBracket);

    IF (Scanner.lookahead = TokenType.Colon) THEN
      Scanner.Match(TokenType.Colon, TokenType.Colon);
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      WHILE (Scanner.lookahead = TokenType.Period) DO
        Scanner.Match(TokenType.Period, TokenType.Period);
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      END;
    END;
  END ParseFormalParameters;


PROCEDURE ParseFPSection () RAISES {Scanner.IOError, Scanner.ScanError} =
  BEGIN
    IF (Scanner.lookahead = TokenType.Var) THEN
      Scanner.Match(TokenType.Var, TokenType.Var);
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    ELSE
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    END;
    WHILE (Scanner.lookahead = TokenType.Comma) DO
      Scanner.Match(TokenType.Comma, TokenType.Comma);
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    END;
    Scanner.Match(TokenType.Colon, TokenType.Colon);
    ParseFormType();
  END ParseFPSection;


PROCEDURE ParseFormType () RAISES {Scanner.IOError, Scanner.ScanError} =
  BEGIN
    IF (Scanner.lookahead = TokenType.Array) THEN
      Scanner.Match(TokenType.Array, TokenType.Array);
      Scanner.Match(TokenType.Of, TokenType.Of);
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      WHILE (Scanner.lookahead = TokenType.Period) DO
        Scanner.Match(TokenType.Period, TokenType.Period);
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      END;
    ELSE
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      WHILE (Scanner.lookahead = TokenType.Period) DO
        Scanner.Match(TokenType.Period, TokenType.Period);
        Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      END;
    END;
  END ParseFormType;


PROCEDURE ParseModuleDeclaration ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Module, TokenType.Module);
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    IF (Scanner.lookahead = TokenType.ALBracket) THEN
      Scanner.Match(TokenType.ALBracket, TokenType.ALBracket);
      ParseConstExpression();
      Scanner.Match(TokenType.ARBracket, TokenType.ARBracket);
    END;
    Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);

    WHILE ((Scanner.lookahead = TokenType.Import)
             OR (Scanner.lookahead = TokenType.From)) DO
      ParseImport();
    END;
    IF (Scanner.lookahead = TokenType.Export) THEN ParseExport(); END;

    ParseBlock();
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
  END ParseModuleDeclaration;


PROCEDURE ParseExport () RAISES {Scanner.IOError, Scanner.ScanError} =
  BEGIN
    ErrorOut("warning: EXPORT clause ommited");
    Scanner.Match(TokenType.Export, TokenType.EmptyString);
    IF (Scanner.lookahead = TokenType.Qualified) THEN
      Scanner.Match(TokenType.Qualified, TokenType.EmptyString);
    END;
    Scanner.Match(TokenType.Identifier, TokenType.EmptyString);
    WHILE (Scanner.lookahead = TokenType.Comma) DO
      Scanner.Match(TokenType.Comma, TokenType.EmptyString);
      Scanner.Match(TokenType.Identifier, TokenType.EmptyString);
    END;
    Scanner.Match(TokenType.Semicolon, TokenType.EmptyString);
  END ParseExport;


PROCEDURE ParseImport () RAISES {Scanner.IOError, Scanner.ScanError} =
  BEGIN
    IF (Scanner.lookahead = TokenType.From) THEN
      Scanner.Match(TokenType.From, TokenType.From);
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
      Scanner.Match(TokenType.Import, TokenType.Import);
    ELSE
      Scanner.Match(TokenType.Import, TokenType.Import);
    END;

    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    WHILE (Scanner.lookahead = TokenType.Comma) DO
      Scanner.Match(TokenType.Comma, TokenType.Comma);
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    END;
    Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
  END ParseImport;


PROCEDURE ParseDefinitionModule ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Match(TokenType.Definition, TokenType.Interface);
    IF (Scanner.lookahead = TokenType.For) THEN
      Scanner.Match(TokenType.For, TokenType.For);
      Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    END;
    Scanner.Match(TokenType.Module, TokenType.EmptyString);
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);

    WHILE ((Scanner.lookahead = TokenType.Import)
             OR (Scanner.lookahead = TokenType.From)) DO
      ParseImport();
    END;
    WHILE (Scanner.lookahead = TokenType.Export) DO ParseExport(); END;

    WHILE ((Scanner.lookahead = TokenType.Const)
             OR (Scanner.lookahead = TokenType.Type)
             OR (Scanner.lookahead = TokenType.Var)
             OR (Scanner.lookahead = TokenType.Procedure)) DO
      ParseDefinition();
    END;

    Scanner.Match(TokenType.End, TokenType.End);
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    Scanner.Match(TokenType.Period, TokenType.Period);
  END ParseDefinitionModule;


PROCEDURE ParseDefinition ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    CASE Scanner.lookahead OF
      TokenType.Const =>
        Scanner.Match(TokenType.Const, TokenType.Const);
        WHILE (Scanner.lookahead = TokenType.Identifier) DO
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          Scanner.Match(TokenType.Equal, TokenType.Equal);
          ParseConstExpression();
          Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
        END;
    | TokenType.Type =>
        Scanner.Match(TokenType.Type, TokenType.Type);
        WHILE (Scanner.lookahead = TokenType.Identifier) DO
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          IF (Scanner.lookahead = TokenType.Equal) THEN
            Scanner.Match(TokenType.Equal, TokenType.Equal);
            ParseType();
          END;
          Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
        END;
    | TokenType.Var =>
        Scanner.Match(TokenType.Var, TokenType.Var);
        WHILE (Scanner.lookahead = TokenType.Identifier) DO
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          WHILE (Scanner.lookahead = TokenType.Comma) DO
            Scanner.Match(TokenType.Comma, TokenType.Comma);
            Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          END;
          Scanner.Match(TokenType.Colon, TokenType.Colon);
          ParseType();
          Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
        END;
    | TokenType.Procedure =>
        ParseProcedureHeading();
        Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
    ELSE
      RAISE ParserError;
    END;
  END ParseDefinition;


PROCEDURE ParseDeclaration ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    CASE Scanner.lookahead OF
      TokenType.Const =>
        Scanner.Match(TokenType.Const, TokenType.Const);
        WHILE (Scanner.lookahead = TokenType.Identifier) DO
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          Scanner.Match(TokenType.Equal, TokenType.Equal);
          ParseConstExpression();
          Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
        END;
    | TokenType.Type =>
        Scanner.Match(TokenType.Type, TokenType.Type);
        WHILE (Scanner.lookahead = TokenType.Identifier) DO
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          Scanner.Match(TokenType.Equal, TokenType.Equal);
          ParseType();
          Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
        END;
    | TokenType.Var =>
        Scanner.Match(TokenType.Var, TokenType.Var);
        WHILE (Scanner.lookahead = TokenType.Identifier) DO
          Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          WHILE (Scanner.lookahead = TokenType.Comma) DO
            Scanner.Match(TokenType.Comma, TokenType.Comma);
            Scanner.Match(TokenType.Identifier, TokenType.Identifier);
          END;
          Scanner.Match(TokenType.Colon, TokenType.Colon);
          ParseType();
          Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
        END;
    | TokenType.Procedure =>
        ParseProcedureDeclaration();
        Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
    | TokenType.Module =>
        ErrorOut("warning: local MODULE");
        ParseModuleDeclaration();
        Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);
    ELSE
    END;
  END ParseDeclaration;


PROCEDURE ParseProgramModule ()
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    IF (Scanner.lookahead = TokenType.Implementation) THEN
      Scanner.Match(Scanner.lookahead, TokenType.EmptyString);
      Scanner.Match(TokenType.Module, TokenType.Module);
    ELSE
      Scanner.Match(TokenType.Module, TokenType.Module);
    END;
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    Scanner.Match(TokenType.Semicolon, TokenType.Semicolon);

    WHILE ((Scanner.lookahead = TokenType.Import)
             OR (Scanner.lookahead = TokenType.From)) DO
      ParseImport();
    END;
    ParseBlock();
    Scanner.Match(TokenType.Identifier, TokenType.Identifier);
    Scanner.Match(TokenType.Period, TokenType.Period);
  END ParseProgramModule;


PROCEDURE ParseCompilationUnit (name: TEXT)
  RAISES {Scanner.IOError, Scanner.ScanError, ParserError} =
  BEGIN
    Scanner.Init(name);
    CASE Scanner.lookahead OF
      TokenType.Definition => ParseDefinitionModule();
    | TokenType.Implementation => ParseProgramModule();
    | TokenType.Module => ParseProgramModule();
    | TokenType.Procedure => ParseProcedureDeclaration();
    ELSE
      ParseStatementSequence();
    END;
    Scanner.Match(TokenType.NoToken, TokenType.NoToken);
    Scanner.Exit();
  END ParseCompilationUnit;


PROCEDURE Check (name: TEXT; VAR ok: BOOLEAN) =
  BEGIN
    ok := TRUE;
    fileName := name;

    TRY
      ParseCompilationUnit(name);
    EXCEPT
      Scanner.IOError (m) => ErrorOut(m); ok := FALSE;
    | Scanner.ScanError => ErrorOut("scan error"); ok := FALSE;
    | ParserError => ErrorOut("parse error"); ok := FALSE;
    END;
  END Check;


PROCEDURE ErrorOut (message: TEXT) =
  <* FATAL Wr.Failure, Thread.Alerted *>

  BEGIN
    Wr.PutText(Stdio.stderr, Fmt.F("\"%s\", line %s: %s\n", fileName,
                                   Fmt.Int(Scanner.line), message))
  END ErrorOut;

BEGIN
END Parser.
