(**********************************************************************)
(** - Copyright (C) 1991 Lehrstuhl III fuer Informatik, RWTH Aachen - *)
(** This library is free software; you can redistribute it and/or     *)
(** modify it under the terms of the GNU Library General Public       *)
(** License as published by the Free Software Foundation; either      *)
(** version 2 of the License, or (at your option) any later version.  *)
(** See the file COPYING.LIB for details.                             *)
(**                                                                   *)
(** Author of the generic parser: Frank Hoefer                        *)
(** For the Modula-2 to -3 converter: Peter Klein                     *)
(**********************************************************************)
MODULE Scanner;

IMPORT Wr, Rd, FileStream, Stdio, Thread, Char, Text, Fmt;


CONST HashTableSize = 97;


TYPE
  HashTableIndex = [0 .. HashTableSize - 1];
  HashTableEntry = RECORD
                     used : BOOLEAN;
                     id   : TEXT;
                     token: TokenType;
                   END;


VAR
  scanFile      : Rd.T;
  table         : ARRAY HashTableIndex OF HashTableEntry;
  lastReadWasEOL: BOOLEAN;
  ch            : CHAR;


PROCEDURE InitHashTable () =
  BEGIN
    FOR i := 0 TO HashTableSize - 1 DO
      table[i].used := FALSE;
      table[i].id := "";
    END;
    AddToHash("AND", TokenType.And);
    AddToHash("ARRAY", TokenType.Array);
    AddToHash("BEGIN", TokenType.Begin);
    AddToHash("BY", TokenType.By);
    AddToHash("CASE", TokenType.Case);
    AddToHash("CONST", TokenType.Const);
    AddToHash("DEFINITION", TokenType.Definition);
    AddToHash("DIV", TokenType.Div);
    AddToHash("DO", TokenType.Do);
    AddToHash("ELSE", TokenType.Else);
    AddToHash("ELSIF", TokenType.Elsif);
    AddToHash("END", TokenType.End);
    AddToHash("EXIT", TokenType.Exit);
    AddToHash("EXPORT", TokenType.Export);
    AddToHash("FOR", TokenType.For);
    AddToHash("FORWARD", TokenType.Forward);
    AddToHash("FROM", TokenType.From);
    AddToHash("IF", TokenType.If);
    AddToHash("IMPLEMENTATION", TokenType.Implementation);
    AddToHash("IMPORT", TokenType.Import);
    AddToHash("IN", TokenType.In);
    AddToHash("LOOP", TokenType.Loop);
    AddToHash("MOD", TokenType.Mod);
    AddToHash("MODULE", TokenType.Module);
    AddToHash("NOT", TokenType.Not);
    AddToHash("OF", TokenType.Of);
    AddToHash("OR", TokenType.Or);
    AddToHash("POINTER", TokenType.Pointer);
    AddToHash("PROCEDURE", TokenType.Procedure);
    AddToHash("QUALIFIED", TokenType.Qualified);
    AddToHash("REM", TokenType.Rem);
    AddToHash("RECORD", TokenType.Record);
    AddToHash("REPEAT", TokenType.Repeat);
    AddToHash("RETURN", TokenType.Return);
    AddToHash("SET", TokenType.Set);
    AddToHash("THEN", TokenType.Then);
    AddToHash("TO", TokenType.To);
    AddToHash("TYPE", TokenType.Type);
    AddToHash("UNTIL", TokenType.Until);
    AddToHash("VAR", TokenType.Var);
    AddToHash("WHILE", TokenType.While);
    AddToHash("WITH", TokenType.With);
  END InitHashTable;



PROCEDURE AddToHash (s: TEXT; t: TokenType) =
  VAR h, d: HashTableIndex;

  BEGIN
    h := 0;
    FOR i := 0 TO Text.Length(s) - 1 DO
      h := (2 * h + ORD(Text.GetChar(s, i))) MOD HashTableSize;
    END;

    d := 1;
    WHILE table[h].used DO h := (h + d) MOD HashTableSize; d := d + 2; END;

    table[h] := HashTableEntry{used := TRUE, id := s, token := t};
  END AddToHash;


PROCEDURE TestReservedWord () =
  VAR h, d: HashTableIndex;

  BEGIN
    h := 0;
    FOR i := 0 TO Text.Length(name) - 1 DO
      h := (2 * h + ORD(Text.GetChar(name, i))) MOD HashTableSize;
    END;

    d := 1;
    WHILE (table[h].used AND (NOT Text.Equal(name, table[h].id))) DO
      h := (h + d) MOD HashTableSize;
      d := d + 2;
    END;

    IF table[h].used THEN
      lookahead := table[h].token;
    ELSE
      lookahead := TokenType.Identifier;
    END;
  END TestReservedWord;


(*******************************************************************)
PROCEDURE Init (name: TEXT) RAISES {IOError} =
  BEGIN
    IF (Text.Length(name) > 0) THEN
      TRY
        scanFile := FileStream.OpenRead(name);
      EXCEPT
        Rd.Failure => RAISE IOError("error in opening input file");
      END;
    ELSE
      scanFile := Stdio.stdin;
    END;

    line := 1;
    lastReadWasEOL := TRUE;
    Read();
    GetNextToken();
  END Init;


PROCEDURE Exit () RAISES {IOError} =
  BEGIN
    TRY
      Rd.Close(scanFile);
    EXCEPT
      Rd.Failure, Thread.Alerted =>
        RAISE IOError("error in closing input file");
    END;
  END Exit;


PROCEDURE Read () RAISES {IOError} =
  BEGIN
    TRY
      IF (lastReadWasEOL) THEN lastReadWasEOL := FALSE; END;
      IF Rd.EOF(scanFile) THEN
        ch := Char.NUL;
      ELSE
        ch := Rd.GetChar(scanFile);
      END;

      IF (ch = Char.NL) THEN lastReadWasEOL := TRUE; INC(line); END;
    EXCEPT
      Rd.Failure, Thread.Alerted, Rd.EndOfFile =>
        RAISE IOError("error in reading input file");
    END;
  END Read;


(********************************************************************)
PROCEDURE GetNextToken () RAISES {IOError} =
  VAR
    comp        : CHAR;
    commentLevel: CARDINAL;
    comment     : TEXT;
    l           : CARDINAL;

  BEGIN
    TRY
      WHILE ((ch = Char.SP) OR (ch = Char.NL) OR (ch = Char.HT)) DO
        Wr.PutChar(Stdio.stdout, ch);
        Read();
      END;

      REPEAT
        WHILE ((ch = Char.SP) OR (ch = Char.NL) OR (ch = Char.HT)) DO
          Wr.PutChar(Stdio.stdout, ch);
          Read();
        END;

        CASE ch OF
          '+' => lookahead := TokenType.Plus; Read();
        | '-' => lookahead := TokenType.Minus; Read();
        | '*' => lookahead := TokenType.Mult; Read();
        | '/' => lookahead := TokenType.Division; Read();
        | ':' =>
            Read();
            IF (ch = '=') THEN
              lookahead := TokenType.Assign;
              Read();
            ELSE
              lookahead := TokenType.Colon;
            END;
        | '&' => lookahead := TokenType.Ampersand; Read();
        | '.' =>
            Read();
            IF (ch = '.') THEN
              lookahead := TokenType.Range;
              Read();
            ELSE
              lookahead := TokenType.Period;
            END;
        | ',' => lookahead := TokenType.Comma; Read();
        | ';' => lookahead := TokenType.Semicolon; Read();
        | '(' =>
            comment := "(";
            Read();
            IF (ch = '*') THEN
              commentLevel := 1;
              comment := comment & "*";
              Read();
              WHILE (commentLevel > 0) DO
                IF (ch = '*') THEN
                  comment := comment & "*";
                  Read();
                  IF (ch = ')') THEN
                    comment := comment & ")";
                    Read();
                    DEC(commentLevel);
                    IF (commentLevel = 0) THEN
                      lookahead := TokenType.Comment;
                      Wr.PutText(Stdio.stdout, comment);
                    END;
                  END;
                ELSIF (ch = '(') THEN
                  comment := comment & "(";
                  Read();
                  IF (ch = '*') THEN
                    INC(commentLevel);
                    comment := comment & "*";
                    Read();
                  END;
                ELSIF ((ch = Char.HT) OR (ch = Char.NL) OR (ch >= Char.SP)) THEN
                  comment := comment & Text.FromChar(ch);
                  Read();
                ELSE
                  commentLevel := 0;
                  lookahead := TokenType.BadToken;
                END;
              END;
            ELSE
              lookahead := TokenType.LBracket;
            END;
        | ')' => lookahead := TokenType.RBracket; Read();
        | '[' => lookahead := TokenType.ALBracket; Read();
        | ']' => lookahead := TokenType.ARBracket; Read();
        | '{' => lookahead := TokenType.CLBracket; Read();
        | '}' => lookahead := TokenType.CRBracket; Read();
        | '^' => lookahead := TokenType.Arrow; Read();
        | '~' => lookahead := TokenType.Not; Read();
        | '=' => lookahead := TokenType.Equal; Read();
        | '#' => lookahead := TokenType.NEqual; Read();
        | '<' =>
            Read();
            IF (ch = '=') THEN
              lookahead := TokenType.LEqual;
              Read();
            ELSIF (ch = '>') THEN
              lookahead := TokenType.NEqual;
              Read();
            ELSE
              lookahead := TokenType.Less;
            END;
        | '>' =>
            Read();
            IF (ch = '=') THEN
              lookahead := TokenType.GEqual;
              Read();
            ELSE
              lookahead := TokenType.Greater;
            END;
        | '|' => lookahead := TokenType.Bar; Read();
        | '"', '\'' =>
            comp := ch;
            name := Text.FromChar(comp);
            Read();
            WHILE (NOT Rd.EOF(scanFile) AND (ch # comp) AND (ch >= Char.SP)) DO
              name := name & Text.FromChar(ch);
              Read();
            END;
            IF (NOT Rd.EOF(scanFile) AND (ch = comp)) THEN
              lookahead := TokenType.StringConst;
              name := name & Text.FromChar(comp);
              Read();
            ELSE
              lookahead := TokenType.BadToken;
            END;
            l := Text.Length(name);
            IF (((l > 3) OR (l = 2)) AND (Text.GetChar(name, 0) = '\'')) THEN
              name := "\"" & Text.Sub(name, 1, l - 2) & "\"";
            END;
        | 'A' .. 'Z', 'a' .. 'z', '_' =>
            name := Text.FromChar(ch);
            Read();
            WHILE (((ch >= 'A') AND (ch <= 'Z'))
                     OR ((ch >= 'a') AND (ch <= 'z'))
                     OR ((ch >= '0') AND (ch <= '9')) OR (ch = '_')) DO
              name := name & Text.FromChar(ch);
              Read();
            END;
            TestReservedWord();
        | '0' .. '9' =>
            name := Text.FromChar(ch);
            Read();
            WHILE (((ch >= '0') AND (ch <= '9'))
                     OR ((ch >= 'A') AND (ch <= 'H'))) DO
              name := name & Text.FromChar(ch);
              Read();
            END;
            l := Text.Length(name) - 1;
            CASE Text.GetChar(name, l) OF
              'H' => name := "16_" & Text.Sub(name, 0, l);
            | 'B' => name := "8_" & Text.Sub(name, 0, l);
            | 'C' => name := Fmt.F("'\\%03s'", Text.Sub(name, 0, l));
            ELSE
            END;
            IF (ch = '.') THEN
              Read();
              IF (ch = '.') THEN
                Rd.UnGetChar(scanFile);
                lookahead := TokenType.IntegerConst;
              ELSE
                name := name & ".";
                WHILE ((ch >= '0') AND (ch <= '9')) DO
                  name := name & Text.FromChar(ch);
                  Read();
                END;
                IF ((ch = 'e') OR (ch = 'E')) THEN
                  name := name & Text.FromChar(ch);
                  Read();
                  IF ((ch = '+') OR (ch = '-')) THEN
                    name := name & Text.FromChar(ch);
                    Read();
                  END;
                  IF ((ch < '0') OR (ch > '9')) THEN
                    lookahead := TokenType.BadToken;
                  ELSE
                    WHILE ((ch >= '0') AND (ch <= '9')) DO
                      name := name & Text.FromChar(ch);
                      Read();
                    END;
                    lookahead := TokenType.RealConst;
                  END;
                ELSE
                  lookahead := TokenType.RealConst;
                END;
              END;
            ELSE
              lookahead := TokenType.IntegerConst;
            END;
        ELSE
          IF (ch < ' ') THEN
            lookahead := TokenType.NoToken;
          ELSE
            lookahead := TokenType.BadToken;
            Read();
          END;
        END;
      UNTIL (lookahead # TokenType.Comment);
    EXCEPT
      Rd.Failure, Thread.Alerted =>
        RAISE IOError("error in reading input file");
    | Wr.Failure => RAISE IOError("error in writing output");
    END;
  END GetNextToken;


(*********************************************************************)
PROCEDURE Match (wanted, write: TokenType) RAISES {IOError, ScanError} =
  BEGIN
    IF (wanted = lookahead) THEN
      WriteToken(write);
      GetNextToken();
    ELSE
      RAISE ScanError;
    END;
  END Match;


PROCEDURE WriteToken (t: TokenType) RAISES {IOError} =
  BEGIN
    TRY
      CASE t OF
        TokenType.Identifier => Wr.PutText(Stdio.stdout, name);
      | TokenType.StringConst => Wr.PutText(Stdio.stdout, name);
      | TokenType.IntegerConst => Wr.PutText(Stdio.stdout, name);
      | TokenType.RealConst => Wr.PutText(Stdio.stdout, name);
      | TokenType.Plus => Wr.PutText(Stdio.stdout, "+");
      | TokenType.Minus => Wr.PutText(Stdio.stdout, "-");
      | TokenType.Mult => Wr.PutText(Stdio.stdout, "*");
      | TokenType.Division => Wr.PutText(Stdio.stdout, "/");
      | TokenType.Assign => Wr.PutText(Stdio.stdout, ":=");
      | TokenType.Ampersand => Wr.PutText(Stdio.stdout, "&");
      | TokenType.Period => Wr.PutText(Stdio.stdout, ".");
      | TokenType.Comma => Wr.PutText(Stdio.stdout, ",");
      | TokenType.Semicolon => Wr.PutText(Stdio.stdout, ";");
      | TokenType.LBracket => Wr.PutText(Stdio.stdout, "(");
      | TokenType.ALBracket => Wr.PutText(Stdio.stdout, "[");
      | TokenType.CLBracket => Wr.PutText(Stdio.stdout, "{");
      | TokenType.RBracket => Wr.PutText(Stdio.stdout, ")");
      | TokenType.ARBracket => Wr.PutText(Stdio.stdout, "]");
      | TokenType.CRBracket => Wr.PutText(Stdio.stdout, "}");
      | TokenType.Arrow => Wr.PutText(Stdio.stdout, "^");
      | TokenType.Equal => Wr.PutText(Stdio.stdout, "=");
      | TokenType.NEqual => Wr.PutText(Stdio.stdout, "#");
      | TokenType.Less => Wr.PutText(Stdio.stdout, "<");
      | TokenType.Greater => Wr.PutText(Stdio.stdout, ">");
      | TokenType.LEqual => Wr.PutText(Stdio.stdout, "<=");
      | TokenType.GEqual => Wr.PutText(Stdio.stdout, ">=");
      | TokenType.Range => Wr.PutText(Stdio.stdout, "..");
      | TokenType.Colon => Wr.PutText(Stdio.stdout, ":");
      | TokenType.Bar => Wr.PutText(Stdio.stdout, "|");
      | TokenType.And => Wr.PutText(Stdio.stdout, "AND");
      | TokenType.Array => Wr.PutText(Stdio.stdout, "ARRAY");
      | TokenType.Begin => Wr.PutText(Stdio.stdout, "BEGIN");
      | TokenType.By => Wr.PutText(Stdio.stdout, "BY");
      | TokenType.Case => Wr.PutText(Stdio.stdout, "CASE");
      | TokenType.Const => Wr.PutText(Stdio.stdout, "CONST");
      | TokenType.Definition => Wr.PutText(Stdio.stdout, "DEFINITION");
      | TokenType.Div => Wr.PutText(Stdio.stdout, "DIV");
      | TokenType.Do => Wr.PutText(Stdio.stdout, "DO");
      | TokenType.Else => Wr.PutText(Stdio.stdout, "ELSE");
      | TokenType.Elsif => Wr.PutText(Stdio.stdout, "ELSIF");
      | TokenType.End => Wr.PutText(Stdio.stdout, "END");
      | TokenType.Exit => Wr.PutText(Stdio.stdout, "EXIT");
      | TokenType.Export => Wr.PutText(Stdio.stdout, "EXPORT");
      | TokenType.For => Wr.PutText(Stdio.stdout, "FOR");
      | TokenType.Forward => Wr.PutText(Stdio.stdout, "FORWARD");
      | TokenType.From => Wr.PutText(Stdio.stdout, "FROM");
      | TokenType.If => Wr.PutText(Stdio.stdout, "IF");
      | TokenType.Implementation =>
          Wr.PutText(Stdio.stdout, "IMPLEMENTATION");
      | TokenType.Import => Wr.PutText(Stdio.stdout, "IMPORT");
      | TokenType.In => Wr.PutText(Stdio.stdout, "IN");
      | TokenType.Loop => Wr.PutText(Stdio.stdout, "LOOP");
      | TokenType.Mod => Wr.PutText(Stdio.stdout, "MOD");
      | TokenType.Module => Wr.PutText(Stdio.stdout, "MODULE");
      | TokenType.Not => Wr.PutText(Stdio.stdout, "NOT");
      | TokenType.Of => Wr.PutText(Stdio.stdout, "OF");
      | TokenType.Or => Wr.PutText(Stdio.stdout, "OR");
      | TokenType.Pointer => Wr.PutText(Stdio.stdout, "POINTER");
      | TokenType.Procedure => Wr.PutText(Stdio.stdout, "PROCEDURE");
      | TokenType.Qualified => Wr.PutText(Stdio.stdout, "QUALIFIED");
      | TokenType.Rem => Wr.PutText(Stdio.stdout, "REM");
      | TokenType.Record => Wr.PutText(Stdio.stdout, "RECORD");
      | TokenType.Repeat => Wr.PutText(Stdio.stdout, "REPEAT");
      | TokenType.Return => Wr.PutText(Stdio.stdout, "RETURN");
      | TokenType.Set => Wr.PutText(Stdio.stdout, "SET");
      | TokenType.Then => Wr.PutText(Stdio.stdout, "THEN");
      | TokenType.To => Wr.PutText(Stdio.stdout, "TO");
      | TokenType.Type => Wr.PutText(Stdio.stdout, "TYPE");
      | TokenType.Until => Wr.PutText(Stdio.stdout, "UNTIL");
      | TokenType.Var => Wr.PutText(Stdio.stdout, "VAR");
      | TokenType.While => Wr.PutText(Stdio.stdout, "WHILE");
      | TokenType.With => Wr.PutText(Stdio.stdout, "WITH");

        (* Modula-3 specials: *)
      | TokenType.ImpArrow => Wr.PutText(Stdio.stdout, "=>");
      | TokenType.Branded => Wr.PutText(Stdio.stdout, "BRANDED");

      | TokenType.Ref => Wr.PutText(Stdio.stdout, "REF");
      | TokenType.Interface => Wr.PutText(Stdio.stdout, "INTERFACE");

      | TokenType.EmptyString => Wr.PutText(Stdio.stdout, "");
      ELSE
      END;
    EXCEPT
      Wr.Failure, Thread.Alerted =>
        RAISE IOError("error in writing output");
    END;
  END WriteToken;

BEGIN
  InitHashTable();
END Scanner.
